/**
 * <p>File: EquationSolver.java</p>
 * <p>Title: Numerical methods</p>
 * <p>Description: Course on numerical analysis</p>
 * <p>Date: February 2012</p>
 * @author Faton M. Berisha
 * @version 7.0.0
 */

package numeric;

/**
 * Computes numerical solutions of equations in one variable
 */
public class EquationSolver extends Solver {
	/**
	 * Constructs an equation solver in 1 variable.
	 */
	public EquationSolver() {
		super(1);
	}

	/**
	 * Computes an approximate numerical solution of an equation by Bisection
	 * method.
	 * 
	 * @param f
	 *            function in the equation f(x) = 0
	 * @param a
	 *            left hand side edge of the solution interval
	 * @param b
	 *            right hand side edge of the solution interval
	 * @return the solution
	 */
	public double bisection(Function f, double a, double b) {
		double x = a;
		double fA = f.value(a);
		int i = 0;
		boolean found = false;
		while (!found && i < getNumIterations()) {
			double half = (b - a) / 2;
			x = a + half;
			solution.set(0, 0, x);
			solution.demonstrate(solution);
			double fX = f.value(x);
			if (fX == 0 || half < getTolerance()) {
				found = true;
			} else {
				if ((fA < 0 && fX < 0) || (fA > 0 && fX > 0)) {
					a = x;
					fA = fX;
				} else {
					b = x;
				}
				i++;
			}
		}
		if (solution.isDemo()) {
			solution.deleteObservers();
		}
		if (!found) {
			throw new ArithmeticException("Method failed after "
					+ getNumIterations() + " iterations");
		}
		return x;
	}

	/**
	 * Approximates a fixed point of a function by iterations.
	 * 
	 * @param g
	 *            function in the equation x = g(x)
	 * @param x0
	 *            initial solution
	 * @return the solution
	 */
	public double fixedPoint(Function g, double x0) {
		double x = x0;
		int i = 0;
		solution.set(0, 0, x);
		solution.demonstrate(solution);
		boolean found = false;
		while (!found && i < getNumIterations()) {
			x = g.value(x0);
			solution.set(0, 0, x);
			solution.demonstrate(solution);
			if (Math.abs(x - x0) < getTolerance()) {
				found = true;
			} else {
				i++;
				x0 = x;
			}
		}
		if (solution.isDemo()) {
			solution.deleteObservers();
		}
		if (!found) {
			throw new ArithmeticException("Method failed after "
					+ getNumIterations() + " iterations");
		}
		return x;
	}

	/**
	 * Computes an approximate solution of an equation by Newton's method.
	 * 
	 * @param f
	 *            differentiable function in the equation f(x) = 0
	 * @param x0
	 *            initial solution
	 * @return the solution
	 */
	public double newton(Differentiable f, double x0) {
		double x = x0;
		int i = 0;
		solution.set(0, 0, x);
		solution.demonstrate(solution);
		boolean found = false;
		while (!found && i < getNumIterations()) {
			x = x0 - (f.value(x0) / f.derivative(x0));
			solution.set(0, 0, x);
			solution.demonstrate(solution);
			if (Math.abs(x - x0) < getTolerance()) {
				found = true;
			} else {
				i++;
				x0 = x;
			}
		}
		if (solution.isDemo()) {
			solution.deleteObservers();
		}
		if (!found) {
			throw new ArithmeticException("Method failed after "
					+ getNumIterations() + " iterations");
		}
		return x;
	}

	/**
	 * Computes a solution of an equation by Secant method.
	 * 
	 * @param f
	 *            function in the equation f(x) = 0
	 * @param x0
	 *            first initial solution
	 * @param x1
	 *            second initial solution
	 * @return the solution
	 */
	public double secant(Function f, double x0, double x1) {
		solution.set(0, 0, x0);
		solution.demonstrate(solution);
		double x = x1;
		solution.set(0, 0, x);
		solution.demonstrate(solution);
		double y0 = f.value(x0);
		int i = 2;
		boolean found = false;
		while (!found && i < getNumIterations()) {
			double y1 = f.value(x1);
			x = x1 - y1 * (x1 - x0) / (y1 - y0);
			solution.set(0, 0, x);
			solution.demonstrate(solution);
			if (Math.abs(x - x1) < getTolerance()) {
				found = true;
			} else {
				i++;
				x0 = x1;
				y0 = y1;
				x1 = x;
			}
		}
		if (solution.isDemo()) {
			solution.deleteObservers();
		}
		if (!found) {
			throw new ArithmeticException("Method failed after "
					+ getNumIterations() + " iterations");
		}
		return x;
	}

	/**
	 * Computes a solution of an equation by Regula Falsi method
	 * (method of False Position).
	 * 
	 * @param f
	 *            function in the equation f(x) = 0
	 * @param x0
	 *            first initial solution
	 * @param x1
	 *            second initial solution
	 * @return the solution
	 */
	public double falsi(Function f, double x0, double x1) {
		solution.set(0, 0, x0);
		solution.demonstrate(solution);
		double x = x1;
		solution.set(0, 0, x);
		solution.demonstrate(solution);
		double y0 = f.value(x0);
		double y1 = f.value(x1);
		int i = 2;
		boolean found = false;
		while (!found && i < getNumIterations()) {
			x = x1 - y1 * (x1 - x0) / (y1 - y0);
			solution.set(0, 0, x);
			solution.demonstrate(solution);
			if (Math.abs(x - x1) < getTolerance()) {
				found = true;
			} else {
				i++;
				double y = f.value(x);
				if ((y0 < 0 && y < 0) || (y0 > 0 && y > 0)) {
					x0 = x1;
					y0 = y1;
				}
				x1 = x;
				y1 = y;
			}
		}
		if (solution.isDemo()) {
			solution.deleteObservers();
		}
		if (!found) {
			throw new ArithmeticException("Method failed after "
					+ getNumIterations() + " iterations");
		}
		return x;
	}

	public int getNumVariables() {
		return 1;
	}

	public int getNumEquations() {
		return 1;
	}

}
