/**
 * <p>File: LinearSystemSolver.java</p>
 * <p>Title: Numerical methods</p>
 * <p>Description: Course on numerical analysis</p>
 * <p>Date: December 2011</p>
 * @author Faton M. Berisha
 * @version 6.0.0
 */

package numeric;

import java.util.Observer;

/**
 * Models a system of linear equations.
 */
public class LinearSystemSolver extends Solver {

//	private static final long serialVersionUID = 2;

	private Matrix matrix;

	/**
	 * Creates a linear system with an empty coefficients matrix of
	 * <code>n</code> equations in <code>n</code> unknowns.
	 * 
	 * @param n
	 *            number of equations (and unknowns).
	 */
	public LinearSystemSolver(int n) {
		super(n);
		matrix = new Matrix(n, n + 1);
	}

	/**
	 * Creates a linear system with specified <code>double[][]</code>
	 * coefficients matrix.
	 * 
	 * @param coefficient
	 *            <code>double[][]</code> coefficients matrix
	 */
	public LinearSystemSolver(double[][] coefficient) {
		super(coefficient.length);
		matrix = new Matrix(coefficient);
	}

	/**
	 * Creates a linear system with specified <code>Matrix</code>
	 * coefficients matrix.
	 * 
	 * @param matrix
	 *            <code>Matrix</code> coefficients matrix
	 */
	public LinearSystemSolver(Matrix matrix) {
		super(matrix.getNumRows());
		this.matrix = matrix;
	}

	/**
	 * Performs Gauss elimination to the system matrix.
	 * 
	 * @return the <code>double[]</code> solution
	 */
	public double[] gauss() {
		matrix.gauss();
		return solveUpperTriangular();
	}

	/**
	 * Performs Gauss-Jordan elimination to the system matrix.
	 * 
	 * @return the <code>double[]</code> solution
	 */
	public double[] gaussJordan() {
		matrix.gaussJordan();
		return solveUpperTriangular();
	}

	/**
	 * Solves an upper triangular system.
	 * 
	 * @return the <code>double[]</code> solution
	 */
	public double[] solveUpperTriangular() {
		int n = getNumVariables();
		double[] answer = new double[n];
		for (int i = n - 1; i >= 0; i--) {
			double sum = 0;
			for (int j = i + 1; j < n; j++) {
				sum += matrix.get(i, j) * solution.get(j, 0);
			}
			solution.set(i, 0, (matrix.get(i, n) - sum) / matrix.get(i, i));
		}
		solution.demonstrate(solution);
		for (int i = 0; i < n; i++) {
			answer[i] = solution.get(i, 0);
		}
		return answer;
	}

	/**
	 * Solves the system by Jacobi iterative method.
	 * 
	 * @param initial
	 *            <code>double[]</code> initial solution
	 * @return the <code>double[]</code> approximate solution
	 */
	public double[] jacobi(double[] initial) {
		int n = getNumVariables();
		double[] answer = new double[n];
		Matrix init = new Matrix(initial);
		solution.demonstrate(init);
		int[] rowsPermutation = validRowsPermutation();
		matrix.permutateRows(rowsPermutation);
		boolean found = false;
		int k = 0;
		while (!found && k < getNumIterations()) {
			for (int i = 0; i < n; i++) {
				double sum = 0.0;
				for (int j = 0; j < n; j++) {
					if (i != j) {
						sum += matrix.get(i, j) * init.get(j, 0);
					}
				}
				solution.set(i, 0, (-sum + matrix.get(i, n)) / matrix.get(i, i));
			}
			solution.demonstrate(solution);
			if (solution.subtract(init).norm() < getTolerance()) {
				found = true;
			} else {
				k++;
				try {
					init = (Matrix)solution.clone();
				}
				catch(CloneNotSupportedException e) {}
			}
		}
		if (solution.isDemo()) {
			solution.deleteObservers();
		}
		if (!found) {
			throw new ArithmeticException("Method failed after "
					+ getNumIterations() + " iterations");
		}
		for (int i = 0; i < n; i++) {
			answer[i] = solution.get(i, 0);
		}
		return answer;
	}

	/**
	 * Solves the system by Gauss-Seidel iterative method.
	 * 
	 * @param initial
	 *            <code>double[]</code> initial solution
	 * @return the <code>double[]</code> approximate solution
	 */
	public double[] gaussSeidel(double[] initial) {
		int n = getNumVariables();
		double[] answer = new double[n];
		Matrix init = new Matrix(initial);
		solution.demonstrate(init);
		int[] rowsPermutation = validRowsPermutation();
		matrix.permutateRows(rowsPermutation);
		boolean found = false;
		int k = 0;
		while (!found && k < getNumIterations()) {
			for (int i = 0; i < n; i++) {
				double sum = 0.0;
				for (int j = 0; j < i; j++) {
					sum += matrix.get(i, j) * solution.get(j, 0);
				}
				for (int j = i + 1; j < n; j++) {
					sum += matrix.get(i, j) * init.get(j, 0);
				}
				solution.set(i, 0, (-sum + matrix.get(i, n)) / matrix.get(i, i));
			}
			solution.demonstrate(solution);
			if (solution.subtract(init).norm() < getTolerance()) {
				found = true;
			} else {
				k++;
				try {
					init = (Matrix)solution.clone();
				}
				catch(CloneNotSupportedException e) {}
			}
		}
		if (solution.isDemo()) {
			solution.deleteObservers();
		}
		if (!found) {
			throw new ArithmeticException("Method failed after "
					+ getNumIterations() + " iterations");
		}
		for (int i = 0; i < n; i++) {
			answer[i] = solution.get(i, 0);
		}
		return answer;
	}

	/**
	 * Gets the augmented system matrix.
	 * 
	 * @return <code>Matrix</code> containing augmented system matrix
	 */
	public Matrix getMatrix() {
		Matrix answer = null;
		try {
			answer = (Matrix) matrix.clone();
		} catch (CloneNotSupportedException e) {
		}
		return answer;
	}

	/**
	 * Adds a matrix observer to this linear system solver.
	 * 
	 * @param observer
	 *            the <code>java.util.Observer</code> observer to be added
	 */
	public void addMatrixObserver(Observer observer) {
		matrix.setDemo(true);
		matrix.addObserver(observer);
	}

	public int getNumEquations() {
		return matrix.getNumRows();
	}

	public void setMatrix(Matrix matrix) {
		this.matrix = matrix;
	}
	
	/**
	 * Sorts a given <code>int[]</code> array.
	 * 
	 * @param t
	 *            <code>int[]</code> array to be sorted
	 */
	protected void sort(int[] t) {
		// quickSort(t, 0, t.length - 1);
	}

	/**
	 * Implements recursive Quick Sort for sorting a given <code>int[]</code>
	 * array.
	 * 
	 * @param t
	 *            <code>int[]</code> array to be sorted
	 * @param left
	 *            lower bound of the array index
	 * @param right
	 *            upper bound of the array index
	 */
	protected void quickSort(int[] t, int left, int right) {
		if (right > left) {
			int index = partition(t, left, right);
			quickSort(t, left, index - 1);
			quickSort(t, index + 1, right);
		}
	}

	private int partition(int[] t, int left, int right) {
		int pivot = t[right];
		boolean indicesCross = false;
		int i = left;
		int j = right - 1;
		while (!indicesCross) {
			while (t[i] < pivot) {
				i++;
			}
			while (pivot < t[j] && j > left) {
				j--;
			}
			if (i >= j) {
				indicesCross = true;
			} else {
				int temp = t[i];
				t[i] = t[j];
				t[j] = temp;
				i++;
				j--;
			}
		}
		t[right] = t[i];
		t[i] = pivot;
		return i;
	}

	private int[] validRowsPermutation() {
		int numRows = getNumEquations();
		int[] answer = new int[numRows];
		int[][] valid = new int[numRows][];
		int[] temp = new int[numRows];
		for (int j = 0; j < numRows; j++) {
			int numNonZero = 0;
			for (int i = 0; i < numRows; i++) {
				if (matrix.get(i, j) != 0) {
					temp[numNonZero++] = i;
				}
			}
			if (numNonZero == 0) {
				if (solution.isDemo()) {
					solution.deleteObservers();
				}
				throw new ArithmeticException("Singular matrix");
			}
			valid[j] = new int[numNonZero];
			sort(temp);
			for (int i = 0; i < numNonZero; i++) {
				valid[j][i] = temp[i];
			}
		}
		if (!isValid(0, answer, valid)) {
			if (solution.isDemo()) {
				solution.deleteObservers();
			}
			throw new ArithmeticException("Singular matrix");
		}
		return answer;
	}

	private boolean isValid(int index, int[] permutation, int[][] valid) {
		boolean found = false;
		int k = 0;
		while (!found && k < valid[index].length) {
			boolean repeated = false;
			int ind = 0;
			while (!repeated && ind < index) {
				if (valid[index][k] == permutation[ind]) {
					repeated = true;
				} else {
					ind++;
				}
			}
			if (!repeated) {
				permutation[index] = valid[index][k];
				if (index + 1 < getNumEquations()) {
					found = isValid(index + 1, permutation, valid);
				} else {
					found = true;
				}
			}
			k++;
		}
		return found;
	}

}
