/**
 * <p>File: RadixFloating.java</p>
 * <p>Title: Numerical methods</p>
 * <p>Description: Course on numerical analysis</p>
 * <p>Date: July 2009</p>
 * @author Faton M. Berisha
 * @version 5.0.0
 */

package numeric;

import java.math.RoundingMode;

/**
 * Provides a radix floating point arithmetics with an arbitrary number of
 * significant digits.
 */
public class RadixFloating {

//	private static final long serialVersionUID = 2;

	private static byte defaultRadix = 2;
	private static int defaultNumDigits = 53;
	private static RoundingMode defaultRounding = RoundingMode.HALF_UP;

	private byte radix = defaultRadix;
	private int sign = POSITIVE;
	private int exponent = 0;
	private byte[] mantissa;
	private int numDigits = defaultNumDigits; // Number of significant digits.
	private RoundingMode rounding = defaultRounding; // Rounding mode.

	/**
	 * "Chopping" rounding mode.
	 */
	public static final RoundingMode CHOPPING = RoundingMode.DOWN;

	/**
	 * "Usual" rounding mode.
	 */
	public static final RoundingMode ROUNDING = RoundingMode.HALF_UP;

	/**
	 * Positive sign, indicating <code>this >= 0</code>.
	 */
	public static final int POSITIVE = 0;

	/**
	 * Negative sign, indicating <code>this < 0</code>.
	 */
	public static final int NEGATIVE = 1;

	/**
	 * Converts a <code>double</code> into <code>RadixFloating</code>.
	 * 
	 * @param val
	 *            <code>double</code> value to be converted to
	 *            <code>RadixFloating</code>.
	 */
	public RadixFloating(double val) {
		mantissa = new byte[numDigits];
		if (val != 0) {
			doubleToRadixFloating(val);
		}
	}

	/**
	 * Converts a floating point representation of a number specified by
	 * <code>int[]</code> mantissa, exponent and radix into
	 * <code>RadixFloating</code>.
	 * 
	 * @param mantissa
	 *            <code>int[]</code> specifying mantissa
	 * @param exponent
	 *            <code>int</code> specifying exponent
	 * @param radix
	 *            <code>int</code> specifying radix, assert
	 *            <code>0 < radix < 256</code>.
	 */
	public RadixFloating(int[] mantissa, int exponent, int radix) {
		if (radix < 2 || (byte) radix != radix) {
			throw new IllegalArgumentException("Illegal radix: " + radix);
		}
		this.radix = (byte) radix;
		this.exponent = exponent;
		numDigits = mantissa.length;
		this.mantissa = new byte[numDigits];
		for (int i = 0; i < numDigits; i++) {
			this.mantissa[i] = (byte) mantissa[i];
		}
		normalize();
	}

	/**
	 * Converts this <code>RadixFloating</code> to <code>double</code>.
	 * 
	 * @return this <code>RadixFloating</code> converted to
	 *         <code>double</code>.
	 */
	public double toDouble() {
		double answer = 0;
		for (int i = numDigits - 1; i >= 0; i--) {
			answer += mantissa[i];
			answer /= radix;
		}
		answer *= Math.pow(radix, exponent);
		if (sign == NEGATIVE) {
			answer = -answer;
		}
		return answer;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuffer answer = new StringBuffer();
		if (sign == NEGATIVE) {
			answer.append('-');
		}
		answer.append("0.");
		for (int i = 0; i < numDigits; i++) {
			answer.append(mantissa[i]);
			if (radix > 10) {
				answer.append('-');
			}
		}
		answer.append("E" + exponent);
		answer.append("(" + radix + ")");
		return answer.toString();
	}

	/**
	 * Gets the radix of this <code>RadixFloating</code>.
	 * 
	 * @return radix of this <code>RadixFloating</code>.
	 */
	public byte getRadix() {
		return radix;
	}

	/**
	 * Sets the radix of this <code>RadixFloating</code>.
	 * 
	 * @param radix
	 *            new radix; assert <code>0 < radix < 256</code>.
	 */
	public void setRadix(int radix) {
		if (radix < 2 || (byte) radix != radix) {
			throw new IllegalArgumentException("Illegal radix: " + radix);
		}
		if (radix != this.radix) {
			double val = toDouble();
			this.radix = (byte) radix;
			mantissa = new byte[numDigits];
			if (val != 0) {
				doubleToRadixFloating(val);
			}
		}
	}

	/**
	 * Gets the sign of this <code>RadixFloating</code>.
	 * 
	 * @return sign of this <code>RadixFloating</code>.
	 */
	public int getSign() {
		return sign;
	}

	/**
	 * Sets the sign of this <code>RadixFloating</code>.
	 * 
	 * @param sign
	 *            new sign; assert
	 *            <code>sign == POSITIVE || sign == NEGATIVE</code>.
	 */
	public void setSign(int sign) {
		if (sign != POSITIVE && sign != NEGATIVE) {
			throw new IllegalArgumentException("Illegal sign: " + sign);
		}
		this.sign = sign;
	}

	/**
	 * Gets the number of significant digits of this <code>RadixFloating</code>.
	 * 
	 * @return number of significant digits of this <code>RadixFloating</code>.
	 */
	public int getNumDigits() {
		return numDigits;
	}

	/**
	 * Sets the number of significant digits of this <code>RadixFloating</code>..
	 * 
	 * @param numDigits
	 *            number of significant digits of this
	 *            <code>RadixFloating</code>.
	 */
	public void setNumDigits(int numDigits) {
		if (numDigits <= 0) {
			throw new IllegalArgumentException("Illegal number of digits: "
					+ numDigits);
		}
		if (numDigits > mantissa.length) {
			byte[] tempMantissa = new byte[numDigits];
			for (int i = 0; i < this.numDigits; i++) {
				tempMantissa[i] = mantissa[i];
			}
			mantissa = tempMantissa;
		} else if (numDigits > this.numDigits) {
			for (int i = this.numDigits; i < numDigits; i++) {
				mantissa[i] = 0;
			}
		} else if (numDigits < this.numDigits) {
			roundToDigits(numDigits);
		}
		this.numDigits = numDigits;
	}

	/**
	 * Gets the rounding mode of this <code>RadixFloating</code>.
	 * 
	 * @return rounding mode of this <code>RadixFloating</code>.
	 */
	public RoundingMode getRounding() {
		return rounding;
	}

	/**
	 * Sets the rounding mode of this <code>RadixFloating</code>.
	 * 
	 * @param rounding
	 *            rounding mode of this <code>RadixFloating</code>.
	 */
	public void setRounding(RoundingMode rounding) {
		this.rounding = rounding;
	}

	/**
	 * Gets the default radix of <code>RadixFloating</code>.
	 * 
	 * @return default radix of <code>RadixFloating</code>.
	 */
	public static byte getDefaultRadix() {
		return defaultRadix;
	}

	/**
	 * Sets the default radix of <code>RadixFloating</code>.
	 * 
	 * @param defaultRadix
	 *            new default radix; assert <code>0 < defaultRadix < 256</code>.
	 */
	public static void setDefaultRadix(int defaultRadix) {
		if (defaultRadix < 2 || (byte) defaultRadix != defaultRadix) {
			throw new IllegalArgumentException("Illegal radix: " + defaultRadix);
		}
		RadixFloating.defaultRadix = (byte) defaultRadix;
	}

	/**
	 * Gets the default number of significant digits of
	 * <code>RadixFloating</code>.
	 * 
	 * @return default number of significant digits of
	 *         <code>RadixFloating</code>.
	 */
	public static int getDefaultNumDigits() {
		return defaultNumDigits;
	}

	/**
	 * Sets the default number of significant digits of
	 * <code>RadixFloating</code>.
	 * 
	 * @param numDigits
	 *            default number of significant digits of
	 *            <code>RadixFloating</code>.
	 */
	public static void setDefaultNumDigits(int defaultNumDigits) {
		if (defaultNumDigits <= 0) {
			throw new IllegalArgumentException("Illegal number of digits: "
					+ defaultNumDigits);
		}
		RadixFloating.defaultNumDigits = defaultNumDigits;
	}

	/**
	 * Gets the default rounding mode of <code>RadixFloating</code>.
	 * 
	 * @return default rounding mode of <code>RadixFloating</code>.
	 */
	public static RoundingMode getDefaultRounding() {
		return defaultRounding;
	}

	/**
	 * Sets the default rounding mode of <code>RadixFloating</code>.
	 * 
	 * @param defaultRounding
	 *            default rounding mode of <code>RadixFloating</code>.
	 */
	public static void setDefaultRounding(RoundingMode defaultRounding) {
		RadixFloating.defaultRounding = defaultRounding;
	}

	private void doubleToRadixFloating(double val) {
		if (val < 0) {
			sign = NEGATIVE;
			val = -val;
		}
		exponent = (int) Math.floor(Math.log(val) / Math.log(radix)) + 1;
		val /= Math.pow(radix, exponent);
		int i = 0;
		while (i < numDigits && val > 0) {
			val *= radix;
			mantissa[i] = (byte) val;
			val -= mantissa[i];
			i++;
		}
		if (rounding.equals(RoundingMode.HALF_UP)) {
			if (val >= 0.5) {
				if (sign == POSITIVE || val > 0.5) {
					correctDigit(i - 1, 1);
				}
			}
		}
		// TODO else if...
	}

	private void correctDigit(int index, int difference) {
		int digit = mantissa[index] + difference;
		assert digit > 0 || index > 0 : "Illegal mantissa digit correction";
		if (digit < 0) {
			if (index > 0) {
				mantissa[index] = (byte) (digit + radix);
				correctDigit(index - 1, -1);
			}
		} else if (digit >= radix) {
			mantissa[index] = (byte) (digit - radix);
			if (index > 0) {
				correctDigit(index - 1, 1);
			} else {
				for (int i = numDigits - 1; i > 0; i--) {
					mantissa[i] = mantissa[i - 1];
				}
				mantissa[0] = 1;
				exponent++;
			}
		} else {
			mantissa[index] = (byte) digit;
		}
	}

	private void roundToDigits(int numDigits) {
		assert numDigits > 0 : "Illegal mantissa rounding to digits: "
				+ numDigits;
		if (rounding.equals(RoundingMode.HALF_UP)) {
			if (2 * mantissa[numDigits] >= radix) {
				if (sign == POSITIVE || 2 * mantissa[numDigits] > radix) {
					correctDigit(numDigits - 1, 1);
				}
			}
		}
		// TODO else if...
	}

	private void normalize() {
		boolean found = false;
		int i = 0;
		while (!found && i < numDigits) {
			if (mantissa[i] != 0) {
				found = true;
			} else {
				i++;
			}
		}
		if (found && i > 0) {
			for (int j = i; j < numDigits; j++) {
				mantissa[j - i] = mantissa[j];
			}
			for (int j = numDigits - i; j < numDigits; j++) {
				mantissa[j] = 0;
			}
			exponent -= i;
		}
	}

}
