/**
 * <p>File: SystemSolver.java</p>
 * <p>Title: Numerical methods</p>
 * <p>Description: Course on numerical analysis</p>
 * <p>Date: July 2009</p>
 * @author Faton M. Berisha
 * @version 5.0.0
 */

package numeric;

/**
 * Computes numerical solutions of systems of nonlinear equations.
 */
public class SystemSolver extends Solver {
	/**
	 * Constructs a nonlinear system solver in <code>n</code> variables.
	 * 
	 * @param n
	 *            the number of variables
	 */
	public SystemSolver(int n) {
		super(n);
	}
	
	/**
	 * Approximates a fixed point of a vector function by iterations.
	 * 
	 * @param g
	 *            vector function in the equation x = g(x)
	 * @param initial
	 *            the <code>double[]</code> initial solution
	 * @return the <code>double[]</code> solution
	 */
	public double[] fixedPoint(VectorFunction g, double[] initial) {
		int n = getNumVariables();
		double[] x0 = new double[n];
		for (int i = 0; i < n; i++) {
			x0[i] = initial[i];
		}
		Matrix init = new Matrix(x0);
		double[] x = x0;
		int k = 0;
		solution.demonstrate(init);
		boolean found = false;
		while (!found && k < getNumIterations()) {
			for (int i = 0; i < n; i++) {
				x[i] = g.value(i, x0);
				solution.set(i, 0, x[i]);
			}
			solution.demonstrate(solution);
			if (solution.subtract(init).norm() < getTolerance()) {
				found = true;
			} else {
				k++;
				x0 = x;
				try {
					init = (Matrix)solution.clone();
				}
				catch (CloneNotSupportedException e) {} 
			}
		}
		if (solution.isDemo()) {
			solution.deleteObservers();
		}
		if (!found) {
			throw new ArithmeticException("Method failed after "
					+ getNumIterations() + " iterations");
		}
		return x;
	}

	/**
	 * Computes an approximate solution of a system of nonlinear equations by Newton's method.
	 * 
	 * @param f
	 *            differentiable vector function in the equation f(x) = 0
	 * @param initial
	 *            the <code>double[]</code> initial solution
	 * @return the <code>double[]</code> solution
	 */
	public double[] newton(PartiallyDifferentiable f, double[] initial) {
		int n = getNumVariables();
		double[] x = new double[n];
		for (int i = 0; i < n; i++) {
			x[i] = initial[i];
			solution.set(i, 0, x[i]);
		}
		int k = 0;
		solution.demonstrate(solution);
		boolean found = false;
		while (!found && k < getNumIterations()) {
			// TODO Replace the following 7 lines with
			// Matrix coeff = new Matrix(n, n + 1);
			// ...
			double[][] coeff = new double[n][n+1];
			for (int i = 0; i < n; i++) {
				for (int j = 0; j < n; j++) {
					coeff[i][j] = f.partialDerivative(i, j, x);
				}
				coeff[i][n] = -f.value(i, x);
			}
			
			LinearSystemSolver ls = new LinearSystemSolver(coeff);
			// TODO Replace the following 2 lines with
			// ls.solve();
			ls.gauss();
			//ls.solveUpperTriangular();

			Matrix y = ls.getSolution();
			// TODO Replace the following 3 lines with
			// solution = solution.add(y);
			for (int i = 0; i < n; i++) {
				x[i] = solution.get(i, 0) + y.get(i, 0);
				solution.set(i, 0, x[i]);
			}
			
			solution.demonstrate(solution);
			if (y.norm() < getTolerance()) {
				found = true;
			} else {
				k++;
			}
		}
		if (solution.isDemo()) {
			solution.deleteObservers();
		}
		if (!found) {
			throw new ArithmeticException("Method failed after "
					+ getNumIterations() + " iterations");
		}
		return x;
	}

	public int getNumEquations() {
		return solution.getNumRows();
	}
}
